#ifndef SPM_H
#define SPM_H

/* SPM allows the values of the matrix or vector to be of any data type, including user-created data types such as symbolic expressions and complex numbers. SPM treats the values as opaque buffers, and the user must provide callbacks for operating on the values.

If the values are symbolic expressions, then they will probably need to be freed to avoid memory leaks. SPM allows the user to specify a callback to be called when a value needs to be freed. */

/* spm_value_info is a structure containing callbacks for handling matrix and vector values. It tells SPM how big a buffer to allocate for each value and how to perform basic operations on the values. Each matrix and vector has a spm_value_info associated with it. */
typedef struct
{
	/* The size of a buffer used to store a single value. */
	unsigned int sizeof_value;
	
	/* set_zero_func(ud, buf) should fill 'buf' with the value representing 0. The value placed in 'buf' will "belong to" SPM, which means that SPM will later call free_func() on it. */
	void (*set_zero_func)(void *udata, void *buf);
	
	/* set_one_func(ud, buf) should fill 'buf' with the value representing 1. The value placed in 'buf' will "belong to" SPM. */
	void (*set_one_func)(void *udata, void *buf);
	
	/* is_zero_func(ud, buf) should return 1 if the contents of 'buf' represent the value 0, and 0 otherwise. */
	int  (*is_zero_func)(void *udata, const void* buf);
	
	/* duplicate_func(ud, buf1, buf2) should duplicate the contents of 'buf1' into 'buf2'. If duplicate_func is NULL, then memcpy will be used. */
	void (*duplicate_func)(void *udata, const void *buf1, void *buf2);
	
	/* free_func(ud, buf) should free the contents of the buffer 'buf' if necessary. It should only be necessary to free the contents of the buffer if the matrix values are symbolic expressions or other values with some malloc()ed memory associated with them. If the matrix values are numeric types or simple structures, then 'free_func' can be set to NULL to indicate that nothing needs to be done. */
	void (*free_func)(void *udata, void* buf);
	
	/* An arbitrary pointer that will be passed to the callbacks */
	void *udata;
} spm_value_info;

/* spm_matrix_element is a structure representing a single nonzero element of a matrix. Note that there is no field in spm_matrix_element for the value of the element. This is because the size of the buffer needed to store the value varies. Instead, the value sits in memory immediately after the end of the spm_matrix_element. */
typedef struct _spm_matrix_element spm_matrix_element;
struct _spm_matrix_element
{
	unsigned int row;
	unsigned int col;
	spm_matrix_element *next_in_row;
	spm_matrix_element *next_in_col;
};

/* Given a pointer to a matrix element, SPM_M_VALUE() returns a pointer to that element's value. */
#define SPM_M_VALUE(e) ((void*)((spm_matrix_element*)(e)+1))

/* spm_matrix is a structure representing a sparse matrix. */
typedef struct
{
	/* n_rows and n_columns specify the dimensions of the matrix. */
	int n_rows;
	int n_cols;
	
	/* zero_value is a buffer holding the value for 0. It is created only once, when the matrix is created. This is so that it is not necessary to call set_zero_func() every time a 0 element is returned by spm_matrix_get(). */
	void *zero_value;
	
	/* 'rows' contains pointers to the first element of each row. 'cols' contains pointers to the first element of each column. */
	spm_matrix_element **rows;
	spm_matrix_element **cols;
	
	/* A pointer to the vinfo used for this matrix */
	const spm_value_info *vinfo;
} spm_matrix;

/* spm_vector_element is a structure representing a single nonzero element of a vector. Like spm_matrix_element, the value of spm_vector_element is stored in a buffer after the end of the structure. */
typedef struct _spm_vector_element spm_vector_element;
struct _spm_vector_element
{
	unsigned int index;
	spm_vector_element *next;
};

/* Given a pointer to a vector element, SPM_V_VALUE() returns a pointer to that element's value. */
#define SPM_V_VALUE(e) ((void*)((spm_vector_element*)(e)+1))

/* spm_vector is a structure representing a sparse vector */
typedef struct
{
	/* size is the maximum number of elements that can be in the vector */
	int size;
	
	/* zero_value is a buffer holding the value for 0. It is created only once, when the vector is created. This is so that it is not necessary to call set_zero_func() every time a 0 element is returned by spm_vector_get(). */
	void *zero_value;
	
	/* A pointer to the first element in the vector */
	spm_vector_element *first;
	
	/* A pointer to the vinfo used for this vector */
	const spm_value_info *vinfo;
} spm_vector;

/* spm_matrix_alloc allocates a matrix of size n_rows by n_cols. Its value info will be 'vinfo'. */
spm_matrix *
spm_matrix_alloc(int n_rows, int n_cols, const spm_value_info *vinfo);

/* spm_matrix_free frees a matrix. */
void
spm_matrix_free(spm_matrix *m);

/* spm_matrix_clear sets all of the values of the matrix to 0. */
void
spm_matrix_clear(spm_matrix *m);

/* spm_matrix_identity turns an empty matrix into an identity matrix. If the matrix is not empty, the behavior is undefined. */
void
spm_matrix_identity(spm_matrix *m);

/* spm_matrix_copy duplicates the contents of 'src' into 'dest'. */
void
spm_matrix_copy(spm_matrix *dest, spm_matrix *src);

/* spm_matrix_set copies the value from the buffer 'data' into the matrix at (row,col). The contents of 'data' will "belong to" SPM. */
void
spm_matrix_set(spm_matrix *m, int row, int col, const void *data);

/* spm_matrix_get copies the value from the matrix at (row,col) into the buffer 'data'. The contents of 'data' will still "belong to" SPM, and should not be modified or freed by the user. */
void
spm_matrix_get(spm_matrix *m, int row, int col, void *data);

/* spm_vector_alloc allocates a vector of size 'size'. Its value info will be 'vinfo'. */
spm_vector *
spm_vector_alloc(int size, const spm_value_info *vinfo);

/* spm_vector_free frees a vector. */
void
spm_vector_free(spm_vector *v);

/* spm_vector_set sets the n-th element of the vector to 'data'. The contents of the buffer 'data' will be copied into the vector and will be considered to "belong to" SPM. */
void
spm_vector_set(spm_vector *v, int n, const void *data);

/* spm_vector_get copies the value from the n-th element of the vector into 'data'. The contents of 'data' will still "belong to" SPM, and should not be modified or freed by the user. */
void
spm_vector_get(spm_vector *v, int n, void *data);

#endif
