#include "spm.h"
#include "spm_gaussian.h"
#include "spm_utils.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

/* spm_matrix_gaussian takes liberties with the internal structure of the matrix in order to improve performance, and because the matrix will be destroyed anyway. This happens in two ways:

1. For columns to the right of the pivot column, spm_matrix_swap_rows() and spm_matrix_add_mul_row() do not bother to keep next-in-column pointers in the correct order, because it would take O(n) time to find the correct values for the pointers.

2. For columns to the left of the pivot column, the next-in-column pointers are broken completely, and may lead to segfaults if they are used at all. */

static void
spm_matrix_divide_row(spm_matrix *m, void *augmentation, const spm_gaussian_cbs *cbs, const spm_gaussian_info *ginfo, int row, void *buf_denom)
{
	/* buf_denom is allowed to be an element of the row being divided. If buf_denom isn't one of the value buffers of an element of the row being divided, though, then it should be completely independent from the row. */
	
	int divide_the_denominator_itself = 0;
	
	spm_matrix_element *e;
	for (e=m->rows[row]; e; e=e->next_in_row)
	{
		if (SPM_M_VALUE(e) == buf_denom)
		{
			/* We can't perform the division right now, because that would modify our denominator. Store it for later instead. */
			
			assert(!divide_the_denominator_itself); /* We should only reach this point once */
			divide_the_denominator_itself = 1;
		}
		else
		{
			ginfo->divide_func(ginfo->udata, SPM_M_VALUE(e), buf_denom, SPM_M_VALUE(e));
		}
	}
	
	cbs->divide_cb(cbs->udata, augmentation, row, buf_denom);
	
	/* Now go back and divide our denominator if necessary. Since our denominator is being divided by itself, the result will always be 1. */
	if (divide_the_denominator_itself)
	{
		m->vinfo->set_one_func(m->vinfo->udata, buf_denom);
	}
}

static void 
spm_matrix_swap_rows(spm_matrix *m, void *augmentation, const spm_gaussian_cbs *cbs, int row1, int row2)
{
	if (row1==row2) return; /* Nothing to be done */
	
	spm_matrix_element *e;
	
	for (e=m->rows[row1]; e; e=e->next_in_row)
	{
		e->row = row2;
	}
	
	for (e=m->rows[row2]; e; e=e->next_in_row)
	{
		e->row = row1;
	}
	
	spm_matrix_element *temp;
	temp          = m->rows[row1];
	m->rows[row1] = m->rows[row2];
	m->rows[row2] = temp;
	
	/* Don't bother correcting the next-in-column pointers */
	
	cbs->swap_cb(cbs->udata, augmentation, row1, row2);
}

static void
spm_matrix_sub_mul_row(spm_matrix *m, void *augmentation, const spm_gaussian_cbs *cbs, const spm_gaussian_info *ginfo, spm_matrix_element *pivot, int row1, spm_matrix_element *mul_element, int row2)
{
	/* spm_matrix_sub_mul_row subtracts row1 * mul_element from row2. mul_element, which would become 0, is destroyed in the process.
	
	pivot must be an element of row1.
	mul_element must be an element of row2.
	pivot->col must be the same as mul_element->col.
	pivot must be equal to 1.
	pivot must be the first nonzero element of row1.
	mul_element must be the first or second nonzero element of row2.
	*/
	
	assert(pivot->row == row1);
	assert(pivot->col == mul_element->col);
	assert(mul_element->row == row2);
	assert(m->rows[row1] == pivot);

	void *scratch1 = malloc(m->vinfo->sizeof_value);
	
	spm_matrix_element *e1, **e2_slot;
	e2_slot = &mul_element->next_in_row;
	for (e1=pivot->next_in_row; e1; e1=e1->next_in_row)
	{
		ginfo->multiply_func(ginfo->udata, SPM_M_VALUE(mul_element), SPM_M_VALUE(e1), scratch1);
		
		/* Find the slot in row2 that corresponds to the element in row1 */
		while (*e2_slot && (*e2_slot)->col < e1->col)
		{
			e2_slot = &(*e2_slot)->next_in_row;
		}
		
		/* Create a new element if necessary */
		if (!*e2_slot || (*e2_slot)->col > e1->col)
		{
			/* The current value that we are subtracting from to is 0, so there is no element in place already. We must create a new element. */
			
			spm_matrix_element *new = malloc(sizeof(spm_matrix_element)+m->vinfo->sizeof_value);
			new->row = row2;
			new->col = e1->col;
			
			/* Put the element in the correct place in the row, but for speed's sake just stick it at the top of the column. */
			new->next_in_row = *e2_slot;
			new->next_in_col = m->cols[e1->col];
			*e2_slot = new;
			m->cols[e1->col] = new;
			
			/* Initialize the slot to 0 */
			m->vinfo->set_zero_func(m->vinfo->udata, SPM_M_VALUE(*e2_slot));
		}
		
		ginfo->subtract_func(ginfo->udata, SPM_M_VALUE(*e2_slot), scratch1, SPM_M_VALUE(*e2_slot));
		
		/* To be strictly correct, it would be best to check if the new value of *e2_slot is zero,
		and delete the slot if it is. However, if we did that we would have to repair the column
		pointers, which is a pain. Also, it won't do any damage by sitting there with a zero value.
		*/
		
		/* Move on to the next element of row2 */
		e2_slot = &(*e2_slot)->next_in_row;
	}
	
	cbs->subtract_multiply_cb(cbs->udata, augmentation, row1, SPM_M_VALUE(mul_element), row2);
	
	/* mul_element will have become 0 at this point, so just delete it. */
	
	/* Keep the row pointers intact */
	spm_matrix_element **mul_element_slot = &m->rows[row2];
	if (*mul_element_slot != mul_element)
	{
		mul_element_slot = &(*mul_element_slot)->next_in_row;
	}
	assert(*mul_element_slot == mul_element);
	
	*mul_element_slot = mul_element->next_in_row;
	
	/* Don't bother fixing the column pointers, we don't need this column any more */
	
	if (m->vinfo->free_func)
	{
		m->vinfo->free_func(m->vinfo->udata, SPM_M_VALUE(mul_element));
	}
	
	free(mul_element);
}

int
spm_gaussian(spm_matrix *m, const spm_gaussian_info *ginfo, void *augmentation, const spm_gaussian_cbs *cbs)
{
	assert(m->vinfo == ginfo->vinfo);
	assert(m->n_rows >= m->n_cols);
	
	int pivot_col;
	for (pivot_col=0; pivot_col<m->n_cols; pivot_col++)
	{		
		/* Choose a pivot row. */
		spm_matrix_element *pivot=NULL, *candidate;
		int pivot_quality = -1;
		for (candidate=m->cols[pivot_col]; candidate; candidate=candidate->next_in_col)
		{
			if (candidate->row<pivot_col)
			{
				/* We already used this row, can't use it again. */
				continue;
			}
			
			if (ginfo->pivot_check_func)
			{
				int quality = ginfo->pivot_check_func(ginfo->udata, SPM_M_VALUE(candidate));
				if (quality > pivot_quality)
				{
					pivot = candidate;
					pivot_quality = quality;
				}
			}
			else
			{
				/* If no pivot-check function is provided, then use the first nonzero pivot. */
				if (!m->vinfo->is_zero_func(m->vinfo->udata, SPM_M_VALUE(candidate)))
				{
					pivot = candidate;
					break;
				}
			}
		}
		
		if (!pivot)
		{
			/* We ran out of pivot candidates! The matrix is underspecified. See if we can find an empty row to recover with. */
			int empty_row;
			for (empty_row = pivot_col; empty_row<m->n_rows; empty_row++)
			{
				if (!m->rows[empty_row])
				{
					break;
				}
			}
			if (empty_row==m->n_rows)
			{
				/* Can't recover from underspecified matrix because there are no empty rows. */
				spm_matrix_free(m);
				return -1;
			}
			
			/* Fill the augmentation of the empty row */
			if (cbs->unknown_cb)
			{
				int ok = cbs->unknown_cb(cbs->udata, augmentation, pivot_col, empty_row);
				if (!ok)
				{
					/* unknown_cb failed, abort */
					spm_matrix_free(m);
					return -1;
				}
			}
			else
			{
				/* Recovery from underspecified matrices is disabled because there is no unkown_cb specified. */
				spm_matrix_free(m);
				return -1;
			}
			
			/* Insert a value of 1 into the appropriate place in the empty row */
			void *one = malloc(m->vinfo->sizeof_value);
			m->vinfo->set_one_func(m->vinfo->udata, one);
			spm_matrix_set(m, empty_row, pivot_col, one);
			free(one);
			
			/* Proceed as though this row was here all along */
			pivot = m->rows[empty_row];
			assert(pivot->col==pivot_col);
		}
				
		/* Move the row into the next slot, where it belongs */
		spm_matrix_swap_rows(m, augmentation, cbs, pivot->row, pivot_col);
		assert(pivot->row == pivot_col);
				
		/* Divide this row by itself to make the pivot element equal to 1. */
		spm_matrix_divide_row(m, augmentation, cbs, ginfo, pivot->row, SPM_M_VALUE(pivot));
		
		/* Subtract this row from all other rows to make their pivot-col'th element 0 */
		spm_matrix_element *e, *e_next;
		for (e=m->cols[pivot_col]; e; e=e_next)
		{			
			e_next = e->next_in_col; /* Because spm_matrix_sub_mul_row destroys 'e' */
			
			if (e->row == pivot->row)
			{
				/* Don't want to subtract the pivot row from itself */
				continue;
			}
						
			spm_matrix_sub_mul_row(m, augmentation, cbs, ginfo, pivot, pivot->row, e, e->row);
		}
		
		/* Call the debug hook */
		if (cbs->debug_cb)
		{
			cbs->debug_cb(cbs->udata, augmentation);
		}
	}
		
	return 0;
}