#ifndef SPM_GAUSSIAN_H
#define SPM_GAUSSIAN_H

/* spm_gaussian_info describes operations used in the Gaussian elimination of a matrix. */
typedef struct
{
	/* vinfo must be the value info for the matrix to be Gaussian-eliminated. SPM will check to make sure that this is the case. */
	const spm_value_info *vinfo;
	
	/* divide_func(ud, buf_num, buf_denom, buf_out) should divide the contents of buf_num by the contents of buf_denom and store the result in buf_out. buf_num, buf_denom, and/or buf_out may be the same buffer. If the values to be operated on are symbolic expressions, then the function should free the contents of buf_num, but not modify buf_denom. Whatever is placed in buf_out will be considered to "belong to" SPM. */
	void (* divide_func)(void *ud, void *buf_num, void *buf_denom, void *buf_out);
	
	/* multiply_func(ud, buf_in1, buf_in2, buf_out) should multiply the contents of buf_in1 by the contents of buf_in2 and store the result in buf_out. The function should not modify buf_in1 or buf_in2. Whatever is placed in buf_out will be considered to "belong to" SPM. buf_in1, buf_in2, and/or buf_out may be the same buffer. */
	void (* multiply_func)(void *ud, void *buf_in1, void *buf_in2, void *buf_out);
	
	/* subtract_func(ud, buf_in1, buf_in2, buf_out) should subtract the contents of buf_in2 from the contents of buf_in1 and place the result in buf_out. If the values to be operated on are symbolic expressions, then the function should free the contents of buf_in1 and buf_in2. Whatever is placed in buf_out will be considered to "belong to" SPM. buf_in1, buf_in2, and/or buf_out may be the same buffer. */
	void (* subtract_func)(void *ud, void *buf_in1, void *buf_in2, void *buf_out);
	
	/* pivot_check_func(ud, buf) is used to determine whether the value stored in 'buf' would make a good pivot or not. The return value from pivot_check_func should be either -1 or a positive integer. If it is -1, then the value will never be used as a pivot. If it is a positive integer, then the pivot candidate for which pivot_check_func returns the largest integer will be used as the pivot.  */
	int (* pivot_check_func)(void *ud, const void *buf);
	
	/* udata is an arbitrary pointer that will be passed to the functions. */
	void *udata;
} spm_gaussian_info;

/* spm_gaussian_cbs describes callbacks called by Gaussian elimination of a matrix. */
typedef struct
{
	/* divide_cb is a callback that will be called when the given row is being divided by the contents of the buffer 'buf_denom'. It should not modify buf_denom. 'augmentation' will be the same pointer that was passed to spm_matrix_gaussian(). */
	void (* divide_cb)(void *ud, void *augmentation, unsigned int row, void *buf_denom);
	
	/* subtract_multiply_cb is a callback that will be called when the product of row1 and the contents of 'buf_mul' is being subtracted from row2. It should not modify buf_mul. 'augmentation' will be the same pointer that was passed to spm_matrix_gaussian(). */
	void (* subtract_multiply_cb)(void *ud, void *augmentation, unsigned int row1, void *buf_mul, unsigned int row2);
	
	/* swap_cb is a callback that will be called when two rows are being swapped. 'augmentation' will be the same pointer that was passed to spm_matrix_gaussian(). */
	void (* swap_cb)(void *ud, void *augmentation, unsigned int row1, unsigned int row2);
	
	/* unknown_cb is a callback that will be called when the system of equations is underspecified. It indicates that there are an infinite number of possible values for the variable at 'col'. The callback can return 1 or 0. If it returns 1, it should fill the row at 'row' with the value that the variable should be given, and the variable will be assumed to have this value. If it returns 0, then the Gauss-Jordan elimination will be aborted. 'augmentation' will be the same pointer that was passed to spm_matrix_gaussian(). If unknown_cb is NULL, then the Gauss-Jordan elimination will be aborted whenever it is underspecified. */
	int (* unknown_cb)(void *ud, void *augmentation, unsigned int col, unsigned int row);
	
	/* debug_cb is a callback that will be called at the end of each round. It exists to make it
	easier to debug Gauss-Jordan elimination operations. */
	void (* debug_cb)(void *ud, void *augmentation);
	
	/* udata is an arbitrary pointer that will be passed to the callbacks. */
	void *udata;
} spm_gaussian_cbs;

/* spm_gaussianaugmentation_info describes operations that can be used by the callbacks in spm_gaussian_cbs. */
typedef struct
{
	/* vinfo_matrix must be the value info for the matrix being Gaussian-eliminated. */
	const spm_value_info *vinfo_matrix;
	
	/* vinfo_augmentation must be the value info for the augmentation to the matrix. */
	const spm_value_info *vinfo_augmentation;
	
	/* divide_func(ud, buf_num, buf_denom, buf_out) should divide the contents of buf_num by the contents of buf_denom and store the result in buf_out. buf_num and buf_out will be values from the augmentation, but buf_denom will be an element from the matrix being Gaussian-eliminated. buf_num, buf_denom, and/or buf_out may be the same buffer. If the values to be operated on are symbolic expressions, then the function should free the contents of buf_num, but not modify buf_denom. Whatever is placed in buf_out will be considered to "belong to" SPM. */
	void (* divide_func)(void *ud, void *buf_num, void *buf_denom, void *buf_out);
	
	/* multiply_func(ud, buf_in1, buf_in2, buf_out) should multiply the contents of buf_in1 by the contents of buf_in2 and store the result in buf_out. buf_in1 and buf_out will be values from the augmentation, but buf_in2 will be a value from the matrix being Gaussian-eliminated. The function should not modify buf_in1 or buf_in2. Whatever is placed in buf_out will be considered to "belong to" SPM. buf_in1, buf_in2, and/or buf_out may be the same buffer. */
	void (* multiply_func)(void *ud, void *buf_in1, void *buf_in2, void *buf_out);
	
	/* subtract_func(ud, buf_in1, buf_in2, buf_out) should subtract the contents of buf_in2 from the contents of buf_in1 and place the result in buf_out. All three buffers will be values from the augmentation. If the values to be operated on are symbolic expressions, then the function should free the contents of buf_in1 and buf_in2. Whatever is placed in buf_out will be considered to "belong to" SPM. buf_in1, buf_in2, and/or buf_out may be the same buffer. */
	void (* subtract_func)(void *ud, void *buf_in1, void *buf_in2, void *buf_out);
	
	/* udata is an arbitrary pointer that will be passed to the functions. */
	void *udata;
} spm_gaussianaugmentation_info;

/* spm_gaussian performs Gauss-Jordan elimination with partial pivoting on the matrix m. Supposing that 'm' has R rows and C columns, then spm_matrix_gaussian will convert the first C rows of 'm' into an identity matrix by three simple operations:

1. Dividing an entire row by a single value
2. Subtracting a multiple of an entire row from another entire row
3. Swapping two rows

As these operations are being performed on the main matrix, the same operations are being performed on a parallel matrix or vector called the 'augmentation'. Because of the wide variety of possible formats for the augmentation, SPM does not operate on the augmentation itself. Instead, every time spm_gaussian performs one of the three operations, it will call the appropriate callback from 'cbs'. For convenience, the pointer 'augmentation' that is passed to spm_matrix_gaussian will also be passed to the callbacks.

'ginfo' should contain a set of functions for working with the values of the matrix.

spm_gaussian will destroy and free 'm'.

Sometimes the matrix to be Gaussian-eliminated will be 'underspecified' - that is, it does not contain enough equations to find values for all of the variables. If the user of SPM suspects that the matrix may be underspecified, they can pass a fourth callback, 'unknown_cb'. If no value can be found for the variable represented by a column 'col', then SPM will look for an empty row. If an empty row is found, then the col'th element of the row will be set to 1 and unknown_cb will be called to fill the augmentation of that row with whatever value the user wants. In this way, SPM "makes up" new equations on the fly until there are enough equations to find values for all of the variables.

The matrix must have at least as many rows as columns. If the system of equations that you want to solve is underspecified, then you can insert empty rows using spm_matrix_resize until it is square.

If the matrix is underspecified and no empty rows could be found, the return value will be -1. If unknown_cb is NULL or unknown_cb returns 0, the return value will be -1. If the operation is successful, then the return value will be 0. */
int
spm_gaussian(spm_matrix *m,
	const spm_gaussian_info *ginfo,
	void *augmentation,
	const spm_gaussian_cbs *cbs);

/* In many of cases where somebody wants to use spm_gaussian, the augmentation will be in the form of a simple vector of values. spm_gaussian_densevector makes it easy to perform Gaussian elimination with the augmentation as a vector. spm_gaussian_densevector will provide the spm_gaussian_cbs to spm_gaussian; the user must provide a spm_gaussianaugmentation_info object instead to tell spm_gaussian_densevector how to work with the elements of the vector. The size of the vector is assumed to be the same as the number of rows in the matrix. If the matrix is underspecified, then spm_gaussian_densevector will abort. */
int
spm_gaussian_densevector(spm_matrix *m,
	const spm_gaussian_info *ginfo,
	void *vector,
	const spm_gaussianaugmentation_info *gainfo);

/* spm_gaussian_densevector_underspecified is like spm_gaussian_densevector except that it also works with underspecified matrices. Instead of aborting on an underspecified matrix, it calls the
callback 'unknown_cb'. If unknown_cb returns 1, then 'value' is assumed to be filled with the value of the variable numbered 'col'. If unknown_cb returns 0, then the Gauss-Jordan elimination will be aborted. */
int
spm_gaussian_densevector_underspecified(spm_matrix *m,
	const spm_gaussian_info *ginfo,
	void *vector,
	const spm_gaussianaugmentation_info *gainfo,
	int (*unknown_cb)(void *udata, unsigned int col, void *value),
	void *unknown_cb_udata);

/* spm_gaussian_densematrix is like spm_gaussian_densevector, except that the augmentation is a matrix stored in array form. The matrix is stored in row-major form in memory. 'matrix' is a pointer to the first row. 'cols' is the number of elements per row of the matrix. 'stride' is the offset, in bytes, between the start of the each row and the start of the next one. The number of rows in the augmentation matrix is assumed to be the same as the number of rows in 'm'. If the matrix is underspecified, then spm_gaussian_densematrix will abort. */
int
spm_gaussian_densematrix(spm_matrix *m,
	const spm_gaussian_info *ginfo,
	void *augmentation, unsigned int cols, unsigned int stride,
	const spm_gaussianaugmentation_info *gainfo);

/* spm_gaussian_densematrix_underspecified is like spm_gaussian_densematrix except that it does not
abort if the matrix is underspecified. Instead, it calls the callback 'unknown_cb'. unknown_cb can return 0 to abort the Gauss-Jordan elimination, or return 1 and fill the 'value' pointer with the desired value for variable associated with column 'col'. */
int
spm_gaussian_densematrix_underspecified(spm_matrix *m,
	const spm_gaussian_info *ginfo,
	void *augmentation, unsigned int cols, unsigned int stride,
	const spm_gaussianaugmentation_info *gainfo,
	int (* unknown_cb)(void *udata, unsigned int col, void *value),
	void *unknown_cb_udata);

/* spm_gaussian_matrix is like spm_gaussian_vector, except that the augmentation is a sparse matrix. */
int
spm_gaussian_matrix(spm_matrix *m,
	const spm_gaussian_info *ginfo,
	spm_matrix *augmentation,
	const spm_gaussianaugmentation_info *gainfo);

/* spm_gaussian_matrix_underspecified is like spm_gaussian_matrix except that it does not abort if
the matrix is underspecified. Instead, it assigns a value of 0 to one of the 'flexible' variables,
then solves the rest of the system based on that. */
int
spm_gaussian_matrix_underspecified(spm_matrix *m,
	const spm_gaussian_info *ginfo,
	spm_matrix *augmentation,
	const spm_gaussianaugmentation_info *gainfo);

#endif