#include "spm.h"
#include "spm_gaussian.h"
#include <stdlib.h>
#include <string.h>
#include <assert.h>

struct info
{
	unsigned int cols;
	unsigned int stride;
	const spm_gaussianaugmentation_info *gainfo;
	int (*unknown_cb)(void *udata, unsigned int col, void *value);
	void *unknown_cb_udata;
};

static void
spm_dense_divide_cb(void *udata, void *augmentation, unsigned int row, void *buf_denom)
{
	struct info *info = udata;
	const spm_gaussianaugmentation_info *gainfo = info->gainfo;
	int stride = info->stride;
	int cols = info->cols;
	int size = gainfo->vinfo_augmentation->sizeof_value;
	
	void *buf_row = augmentation + stride * row;
	
	int i;
	for (i=0; i<cols; i++)
	{
		void *buf_element = buf_row + i * size;
		gainfo->divide_func(gainfo->udata, buf_element, buf_denom, buf_element);
	}
}

static void
spm_dense_subtract_multiply_cb(void *udata, void *augmentation, unsigned int row1, void *buf_mul, unsigned int row2)
{
	struct info *info = udata;
	const spm_gaussianaugmentation_info *gainfo = info->gainfo;
	int stride = info->stride;
	int cols = info->cols;
	int size = gainfo->vinfo_augmentation->sizeof_value;
	
	void *scratch = malloc(size);
	
	void *buf_row1 = augmentation + stride * row1;
	void *buf_row2 = augmentation + stride * row2;
	
	int i;
	for (i=0; i<cols; i++)
	{
		void *buf_element1 = buf_row1 + i * size;
		void *buf_element2 = buf_row2 + i * size;
		gainfo->multiply_func(gainfo->udata, buf_element1, buf_mul, scratch);
		gainfo->subtract_func(gainfo->udata, buf_element2, scratch, buf_element2);
	}
	
	free(scratch);
}

static void
spm_dense_swap_cb(void *udata, void *augmentation, unsigned int row1, unsigned int row2)
{
	struct info *info = udata;
	const spm_gaussianaugmentation_info *gainfo = info->gainfo;
	int stride = info->stride;
	int cols = info->cols;
	int size = gainfo->vinfo_augmentation->sizeof_value;
	int rowsize = size*cols;
	
	void *scratch = malloc(rowsize);
	void *buf_row1 = augmentation + stride * row1;
	void *buf_row2 = augmentation + stride * row2;
	
	memcpy(scratch, buf_row1, rowsize);
	memcpy(buf_row1, buf_row2, rowsize);
	memcpy(buf_row2, scratch, rowsize);
	
	free(scratch);
}

static int
spm_dense_unknown_cb(void *udata, void *augmentation, unsigned int col, unsigned int row)
{
	struct info *info = udata;
	void *value = augmentation + info->stride * row;
	return info->unknown_cb(info->unknown_cb_udata, col, value);
}

int
spm_gaussian_densematrix(spm_matrix *m,
	const spm_gaussian_info *ginfo,
	void *augmentation, unsigned int cols, unsigned int stride,
	const spm_gaussianaugmentation_info *gainfo)
{
	assert(m->vinfo == gainfo->vinfo_matrix);
	assert(stride >= cols * gainfo->vinfo_augmentation->sizeof_value);
	
	struct info info;
	info.gainfo = gainfo;
	info.stride = stride;
	info.cols = cols;
	
	spm_gaussian_cbs cbs;
	cbs.divide_cb = spm_dense_divide_cb;
	cbs.subtract_multiply_cb = spm_dense_subtract_multiply_cb;
	cbs.swap_cb = spm_dense_swap_cb;
	cbs.unknown_cb = NULL;
	cbs.debug_cb = NULL;
	cbs.udata = &info;
	
	return spm_gaussian(m, ginfo, augmentation, &cbs);
}

int
spm_gaussian_densevector(spm_matrix *m,
	const spm_gaussian_info *ginfo,
	void *vector,
	const spm_gaussianaugmentation_info *gainfo)
{
	return spm_gaussian_densematrix(m, ginfo, vector, 1, gainfo->vinfo_augmentation->sizeof_value, gainfo);
}

int
spm_gaussian_densematrix_underspecified(spm_matrix *m,
	const spm_gaussian_info *ginfo,
	void *augmentation, unsigned int cols, unsigned int stride,
	const spm_gaussianaugmentation_info *gainfo,
	int (*unknown_cb)(void *udata, unsigned int col, void *value),
	void *unknown_cb_udata)
{
	assert(m->vinfo == gainfo->vinfo_matrix);
	assert(stride >= cols * gainfo->vinfo_augmentation->sizeof_value);
	
	struct info info;
	info.gainfo = gainfo;
	info.stride = stride;
	info.cols = cols;
	info.unknown_cb = unknown_cb;
	info.unknown_cb_udata = unknown_cb_udata;
	
	spm_gaussian_cbs cbs;
	cbs.divide_cb = spm_dense_divide_cb;
	cbs.subtract_multiply_cb = spm_dense_subtract_multiply_cb;
	cbs.swap_cb = spm_dense_swap_cb;
	cbs.unknown_cb = spm_dense_unknown_cb;
	cbs.debug_cb = NULL;
	cbs.udata = &info;
	
	return spm_gaussian(m, ginfo, augmentation, &cbs);
}

int
spm_gaussian_densevector_underspecified(spm_matrix *m,
	const spm_gaussian_info *ginfo,
	void *vector,
	const spm_gaussianaugmentation_info *gainfo,
	int (*unknown_cb)(void *udata, unsigned int col, void *value),
	void *unknown_cb_udata)
{
	return spm_gaussian_densematrix_underspecified( m,
													ginfo,
													vector,
													1,
													gainfo->vinfo_augmentation->sizeof_value, gainfo,
													unknown_cb,
													unknown_cb_udata);
}