#include "spm.h"
#include "spm_gaussian.h"
#include "spm_utils.h"
#include <stdlib.h>
#include <string.h>
#include <assert.h>

static void
spm_matrix_divide_cb(void *udata, void *augmentation, unsigned int row, void *buf_denom)
{
	spm_matrix *aug = augmentation;
	spm_gaussianaugmentation_info *gainfo = udata;
	
	spm_matrix_element *e;
	for (e=aug->rows[row]; e; e=e->next_in_row)
	{
		gainfo->divide_func(gainfo->udata, SPM_M_VALUE(e), buf_denom, SPM_M_VALUE(e));
	}
}

static void
spm_matrix_subtract_multiply_cb(void *udata, void *augmentation, unsigned int row1, void *buf_mul, unsigned int row2)
{
	spm_matrix *aug = augmentation;
	spm_gaussianaugmentation_info *gainfo = udata;
	
	void *scratch = malloc(aug->vinfo->sizeof_value);
	
	spm_matrix_element *e1, **e2_slot;
	e2_slot = &aug->rows[row2];
	for (e1=aug->rows[row1]; e1; e1=e1->next_in_row)
	{
		gainfo->multiply_func(gainfo->udata, SPM_M_VALUE(e1), buf_mul, scratch);
		
		/* Find the slot in row2 that corresponds to the element of row1 */
		while (*e2_slot && (*e2_slot)->col<e1->col)
		{
			e2_slot = &(*e2_slot)->next_in_row;
		}
		
		/* Create a new element if necessary */
		if (!*e2_slot || (*e2_slot)->col>e1->col)
		{
			spm_matrix_element *new;
			new = malloc(sizeof(spm_matrix_element)+aug->vinfo->sizeof_value);
			new->row = row2;
			new->col = e1->col;
			
			/* Insert the new element in the correct place in the row, but don't bother fixing the column pointers */
			new->next_in_row = *e2_slot;
			*e2_slot = new;
			
			/* Initially set the element to 0 */
			aug->vinfo->set_zero_func(aug->vinfo->udata, SPM_M_VALUE(*e2_slot));
		}
		
		gainfo->subtract_func(gainfo->udata, SPM_M_VALUE(*e2_slot), scratch, SPM_M_VALUE(*e2_slot));
		
		if (aug->vinfo->is_zero_func(aug->vinfo->udata, SPM_M_VALUE(*e2_slot)))
		{
			/* Delete this row element, because it has become 0 */
			spm_matrix_element *to_delete = *e2_slot;
			*e2_slot = (*e2_slot)->next_in_row;
			
			if (aug->vinfo->free_func)
			{
				aug->vinfo->free_func(aug->vinfo->udata, SPM_M_VALUE(to_delete));
			}
			free(to_delete);
		}
		else
		{
			/* Move on to the next element in the row */
			e2_slot = &(*e2_slot)->next_in_row;
		}
	}
	
	free(scratch);
}

static void
spm_matrix_swap_cb(void *udata, void *augmentation, unsigned int row1, unsigned int row2)
{
	spm_matrix *aug = augmentation;
	
	/* Change row numbers */
	spm_matrix_element *e;
	for (e=aug->rows[row1]; e; e=e->next_in_row)
	{
		e->row = row2;
	}
	for (e=aug->rows[row2]; e; e=e->next_in_row)
	{
		e->row = row1;
	}
	
	/* Swap the actual rows */
	spm_matrix_element *temp;
	temp            = aug->rows[row1];
	aug->rows[row1] = aug->rows[row2];
	aug->rows[row2] = temp;
	
	/* Don't bother updating column pointers; they will be fixed at the end */
}

static int
spm_matrix_unknown_cb(void *udata, void *augmentation, unsigned int col, unsigned int row)
{
	/* What if the augmentation at 'row' is not already empty?
	
	The main matrix at 'row' must be empty, or else spm_matrix_unknown_cb would not have been called. This means that the value of the row must be zero for the situation to make sense. This
	does not mean that the row must be empty. For example, we might have the following system:
	
	[ 1 2 ][ a ]   [ 1 0 ][ c ]
	[     ][   ] = [     ][   ]
	[ 1 2 ][ b ]   [ 0 1 ][ d ]
	
	The solution, of course, is that a and b can be anything and c=d. This system will be reduced to:
	
	[ 1 2 ][ a ]   [ 1 0 ][ c ]
	[     ][   ] = [     ][   ]
	[ 0 0 ][ b ]   [-1 1 ][ d ]
	
	The Gaussian elimination algorithm will then try to use the second row as a underspecification recovery row. The row itself is valid, since (0)a+(0)b = (-1)c+(1)d. The final result will be the same whether or not we clear the row, so it's simpler to just leave it as is.
	
	The only job that this function has is to return 1. */
	
	return 1;
}

int
spm_gaussian_matrix(spm_matrix *m,
	const spm_gaussian_info *ginfo,
	spm_matrix *augmentation,
	const spm_gaussianaugmentation_info *gainfo)
{
	assert(m->vinfo == gainfo->vinfo_matrix);
	assert(augmentation->vinfo == gainfo->vinfo_augmentation);
	
	spm_gaussian_cbs cbs;
	cbs.divide_cb = spm_matrix_divide_cb;
	cbs.subtract_multiply_cb = spm_matrix_subtract_multiply_cb;
	cbs.swap_cb = spm_matrix_swap_cb;
	cbs.unknown_cb = NULL;
	cbs.debug_cb = NULL;
	cbs.udata = (spm_gaussianaugmentation_info *)gainfo; /* The cast is necessary to get rid of const. */
	
	int ok = spm_gaussian(m, ginfo, augmentation, &cbs);
	
	/* Some of the callbacks don't bother updating the row and column pointers of the augmentation matrix. Since we need to preserve the augmentation matrix, we should fix the column pointers. */
	spm_matrix_fix_col_pointers(augmentation);
	
	return ok;
}

int
spm_gaussian_matrix_underspecified(spm_matrix *m,
	const spm_gaussian_info *ginfo,
	spm_matrix *augmentation,
	const spm_gaussianaugmentation_info *gainfo)
{
	assert(m->vinfo == gainfo->vinfo_matrix);
	assert(augmentation->vinfo == gainfo->vinfo_augmentation);
	assert(m->n_rows == augmentation->n_rows);
	
	spm_gaussian_cbs cbs;
	cbs.divide_cb = spm_matrix_divide_cb;
	cbs.subtract_multiply_cb = spm_matrix_subtract_multiply_cb;
	cbs.swap_cb = spm_matrix_swap_cb;
	cbs.unknown_cb = spm_matrix_unknown_cb;
	cbs.debug_cb = NULL;
	cbs.udata = (spm_gaussianaugmentation_info *)gainfo; /* The cast is necessary to get rid of const. */
	
	int ok = spm_gaussian(m, ginfo, augmentation, &cbs);
	
	/* Some of the callbacks don't bother updating the row and column pointers of the augmentation matrix. Since we need to preserve the augmentation matrix, we should fix the column pointers. */
	spm_matrix_fix_col_pointers(augmentation);
	
	return ok;
}
