#include "spm.h"
#include "spm_utils.h"
#include <assert.h>
#include <stdlib.h>
#include <string.h>

spm_matrix *
spm_matrix_alloc(int n_rows, int n_cols, const spm_value_info *vinfo)
{
	spm_matrix *m = malloc(sizeof(spm_matrix));
	if (!m)
	{
		return NULL;
	}
	
	m->n_rows = n_rows;
	m->n_cols = n_cols;
	
	m->vinfo = vinfo;
	
	/* Allocate arrays to hold rows and columns */
	m->rows = malloc(sizeof(void*) * n_rows);
	if (!m->rows)
	{
		free(m);
		return NULL;
	}
	bzero(m->rows, sizeof(void*) * n_rows);
	
	m->cols = malloc(sizeof(void*) * n_cols);
	if (!m->cols)
	{
		free(m->rows);
		free(m);
		return NULL;
	}
	bzero(m->cols, sizeof(void*) * n_cols);
	
	/* Create a prototype zero value */
	m->zero_value = malloc(sizeof(void*) * vinfo->sizeof_value);
	if (!m->zero_value)
	{
		free(m->rows);
		free(m->cols);
		free(m);
		return NULL;
	}
	
	vinfo->set_zero_func(vinfo->udata, m->zero_value);
	
	return m;
}

void
spm_matrix_free(spm_matrix *m)
{
	spm_matrix_clear(m);
	
	free(m->rows);
	free(m->cols);
	
	free(m);
}

void
spm_matrix_clear(spm_matrix *m)
{
	int r;
	for (r=0; r<m->n_rows; r++)
	{
		spm_matrix_element *e, *next;
		for (e=m->rows[r]; e; e=next)
		{
			if (m->vinfo->free_func)
			{
				m->vinfo->free_func(m->vinfo->udata, SPM_M_VALUE(e));
			}
			
			next = e->next_in_row;
			free(e);
		}
	}
	
	bzero(m->rows, sizeof(void*) * m->n_rows);
	bzero(m->cols, sizeof(void*) * m->n_cols);
}

void
spm_matrix_identity(spm_matrix *m)
{
	int n;
	for (n=0; n<m->n_rows && n<m->n_cols; n++)
	{
		spm_matrix_element *new;
		new = malloc(sizeof(spm_matrix_element) + m->vinfo->sizeof_value);
		new->row = new->col = n;
		new->next_in_row = new->next_in_col = NULL;
		
		m->vinfo->set_one_func(m->vinfo->udata, SPM_M_VALUE(new));
		
		m->rows[n] = new;
		m->cols[n] = new;
	}
}

void
spm_matrix_copy(spm_matrix *dest, spm_matrix *src)
{
	assert(src->n_rows == dest->n_rows);
	assert(src->n_cols == dest->n_cols);
	assert(src->vinfo == dest->vinfo);
	
	void *scratch = malloc(dest->vinfo->sizeof_value);
	
	int i;
	for (i=0; i<src->n_rows; i++)
	{
		spm_matrix_element *e;
		for (e=src->rows[i]; e; e=e->next_in_row)
		{
			if (src->vinfo->duplicate_func)
			{
				src->vinfo->duplicate_func(src->vinfo->udata, SPM_M_VALUE(e), scratch);
			}
			else
			{
				memcpy(scratch, SPM_M_VALUE(e), src->vinfo->sizeof_value);
			}
			
			spm_matrix_set(dest, e->row, e->col, scratch);
		}
	}
	
	free(scratch);
}

static void
spm_matrix_set_nonzero(spm_matrix *m, int row, int col, const void *data)
{
	int already_existed;
	spm_matrix_element *e;
	already_existed = spm_matrix_find_or_create_element(m, row, col, &e);
	
	if (already_existed && m->vinfo->free_func)
	{
		m->vinfo->free_func(m->vinfo->udata, SPM_M_VALUE(e));
	}
	
	memcpy(SPM_M_VALUE(e), data, m->vinfo->sizeof_value);
}

static void
spm_matrix_set_zero(spm_matrix *m, int row, int col)
{
	spm_matrix_element **row_slot, **col_slot;
	
	/* Look for an existing element in the row. It it is found, record a pointer to the row-pointer to it. Otherwise, exit, because the value is already zero. */
	row_slot = &(m->rows[row]);
	while(1)
	{
		if (!*row_slot || (*row_slot)->col>col)
		{
			/* There is no element at the given row and column; nothing needs to be done. */
			return;
		}
		else if ((*row_slot)->col==col)
		{
			break;
		}
		else
		{
			row_slot = &(*row_slot)->next_in_row;
		}
	}
	
	/* Find a pointer to the col-pointer to the element */
	col_slot = &(m->cols[col]);
	while(1)
	{
		assert(*col_slot);
		assert((*col_slot)->col<col);
		
		if ((*col_slot)->col==col)
		{
			break;
		}
		else
		{
			col_slot = &(*col_slot)->next_in_col;
		}
	}
	
	assert(*row_slot == *col_slot);
	
	/* Remove the element */
	spm_matrix_element *old = *row_slot;
	*row_slot = old->next_in_row;
	*col_slot = old->next_in_col;
	
	if (m->vinfo->free_func)
	{
		m->vinfo->free_func(m->vinfo->udata, SPM_M_VALUE(old));
	}
	
	free(old);
}

void
spm_matrix_set(spm_matrix *m, int row, int col, const void *data)
{
	assert(m);
	assert(row < m->n_rows);
	assert(col < m->n_cols);
	assert(data);
	
	if (m->vinfo->is_zero_func(m->vinfo->udata, data))
	{
		spm_matrix_set_zero(m, row, col);
	}
	else
	{
		spm_matrix_set_nonzero(m, row, col, data);
	}
}

void
spm_matrix_get(spm_matrix *m, int row, int col, void *data)
{
	spm_matrix_element *e;
	for (e=m->rows[row]; e && e->col<=col; e=e->next_in_row)
	{
		if (e->col==col)
		{
			memcpy(data, SPM_M_VALUE(e), m->vinfo->sizeof_value);
			return;
		}
	}
	
	/* No element was found */
	memcpy(data, m->zero_value, m->vinfo->sizeof_value);
}

/* A version of SPM_M_VALUE in non-macro form, useful in a debugger */
#undef SPM_M_VALUE
void *SPM_M_VALUE(spm_matrix_element *e)
{
	return e+1;
}