#ifndef SPM_MULTIPLY_H
#define SPM_MULTIPLY_H

/* spm_multiply_info describes operations on the values of a matrix during matrix multiplication. It tells SPM how to handle multiplication and addition of matrix elements. */
typedef struct
{
	/* The type of the left-hand matrix to be multiplied. SPM will make sure this is the same as the value info of the first argument to spm_multiply_matrices. */
	const spm_value_info *in1;
	
	/* The type of the right-hand matrix to be multiplied. SPM will make sure this is the same as the value info of the second argument to spm_multiply_matrices. */
	const spm_value_info *in2;
	
	/* The type of the result. SPM will make sure this is the same as the value info of the third argument to spm_multiply_matrices. */
	const spm_value_info *out;
	
	/* multiply_func(ud, buf_in1, buf_in2, buf_out) should multiply the values in buf_in1 and buf_in2 and store the result in buf_out. The value in buf_in1 will be a value from the first input matrix. The value in buf_in2 will be a value from the second input matrix. The result should be a value for the output matrix. The contents of buf_in1 and buf_in2 should not be modified. Whatever is placed in 'buf_out' will "belong to" SPM, which means that SPM will later call free_func() on it. */
	void (*multiply_func)(void *udata, void *buf_in1, void *buf_in2, void *buf_out);
	
	/* add_func(ud, buf_in1, buf_in2, buf_out) should add the values in buf_in1 and buf_in2 and store the result in buf_out. The values in buf_in1 and buf_in2 will both be from the output matrix, and the value placed in buf_out should be for the output matrix. The contents of buf_in1 and buf_in2 can be modified. If buf_in1 and buf_in2 are symbolic expressions, then add_func is expected to free them; SPM will consider them to be destroyed or freed by add_func. The value placed in buf_out is considered to belong to SPM. */
	void (*add_func)(void *udata, void *buf_in1, void *buf_in2, void *buf_out);
	
	/* An arbitrary pointer to be passed to the functions */
	void *udata;
} spm_multiply_info;

/* spm_multiply_matrices multiplies m1 and m2 and stores the result in m_out. It uses the operations described in 'minfo' to add and multiply matrix elements. m1, m2, and m_out can all be of different data types. */
void
spm_multiply_matrices(spm_matrix *m1, spm_matrix *m2, spm_matrix *m_out, const spm_multiply_info *minfo);

/* spm_multiply_matrices_todensematrix is like spm_multiply_matrices except that the result is stored in a dense matrix instead of a sparse matrix. 'm_out' is a pointer to the first element of the first row of the dense matrix. 'stride' is the difference in bytes between the beginning of a row of the dense matrix and the beginning of the next row. The elements of the dense matrix should all be initialized to 0 when it is passed to spm_multiply_matrices_todensematrix. */
void
spm_multiply_matrices_to_densematrix(spm_matrix *m1, spm_matrix *m2, void *m_out, unsigned int stride, const spm_multiply_info *minfo);

#endif