#include "spm.h"
#include "spm_utils.h"
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <stdio.h>

void
spm_matrix_fix_row_pointers(spm_matrix *m)
{
	int row, col;
	
	spm_matrix_element ***ends = malloc(sizeof(void*) * m->n_rows);
	
	for (row=0; row<m->n_rows; row++)
	{
		ends[row] = &m->rows[row];
	}
	
	for (col=0; col<m->n_cols; col++)
	{
		spm_matrix_element *e;
		for (e=m->cols[col]; e; e=e->next_in_col)
		{
			*ends[e->row] = e;
			ends[e->row] = &e->next_in_row;
		}
	}
	
	for (row=0; row<m->n_rows; row++)
	{
		*ends[row] = NULL;
	}
	
	free(ends);
}

void
spm_matrix_fix_col_pointers(spm_matrix *m)
{
	int row, col;
	
	spm_matrix_element ***ends = malloc(sizeof(void*) * m->n_rows);
	
	for (col=0; col<m->n_cols; col++)
	{
		ends[col] = &m->cols[col];
	}
	
	for (row=0; row<m->n_rows; row++)
	{
		spm_matrix_element *e;
		for (e=m->rows[row]; e; e=e->next_in_row)
		{
			*ends[e->col] = e;
			ends[e->col] = &e->next_in_col;
		}
	}
	
	for (col=0; col<m->n_cols; col++)
	{
		*ends[col] = NULL;
	}
	
	free(ends);
}

int
spm_matrix_find_or_create_element(spm_matrix *m, int row, int col, spm_matrix_element **e)
{
	spm_matrix_element **row_slot, **col_slot;
	
	/* Look for an existing element in the row. If there is an existing element, then return it. Otherwise, record a pointer to the row slot that should be filled with the new element. */
	for (row_slot = &m->rows[row];
	     *row_slot && (*row_slot)->col<=col;
	     row_slot = &(*row_slot)->next_in_row)
	{
		if ((*row_slot)->col==col)
		{
			*e = *row_slot;
			return 1;
		}
	}
	
	/* Find the column slot that should be filled with the new element */
	for (col_slot = &m->cols[col];
	     *col_slot && (*col_slot)->row<=row;
	     col_slot = &(*col_slot)->next_in_col)
	{
		continue;
	}
	
	/* Create the new element and put it into the matrix */
	spm_matrix_element *new = malloc(sizeof(spm_matrix_element)+m->vinfo->sizeof_value);
	new->row = row;
	new->col = col;
		
	new->next_in_row = *row_slot;
	*row_slot = new;
	
	new->next_in_col = *col_slot;
	*col_slot = new;
	
	*e = new;
	return 0;
}

int
spm_vector_find_or_create_element(spm_vector *v, int index, spm_vector_element **e)
{
	spm_vector_element **slot;
	
	for (slot=&v->first;
	     *slot && (*slot)->index<=index;
	     slot = &(*slot)->next)
	{
		if ((*slot)->index==index)
		{
			*e = *slot;
			return 1;
		}
	}
	
	spm_vector_element *new = malloc(sizeof(spm_vector_element)+v->vinfo->sizeof_value);
	new->index = index;
	
	new->next = *slot;
	*slot = new;
	
	*e = new;
	return 0;
}

void
spm_matrix_resize(spm_matrix *m, int n_rows, int n_cols)
{
	if (n_rows > m->n_rows)
	{
		/* All that we need to do is expand the row array */
		m->rows = realloc(m->rows, sizeof(void*)*n_rows);
		bzero(m->rows+m->n_rows, sizeof(void*)*(n_rows - m->n_rows));
		
		m->n_rows = n_rows;
	}
	else if (n_rows < m->n_rows)
	{
		/* Cut off the columns at the new boundary */
		int col;
		for (col=0; col<m->n_cols; col++)
		{
			spm_matrix_element *e;
			for (e=m->cols[col]; e; e=e->next_in_col)
			{
				if (e->next_in_col && e->next_in_col->row >= n_rows)
				{
					e->next_in_col = NULL;
					break;
				}
			}
		}
		
		/* Free the rows which no longer exist */
		int row;
		for (row=m->n_rows; row<n_rows; row++)
		{
			spm_matrix_element *e, *e_next;
			for (e=m->rows[row]; e; e=e_next)
			{
				e_next = e->next_in_row;
				
				if (m->vinfo->free_func)
				{
					m->vinfo->free_func(m->vinfo->udata, SPM_M_VALUE(e));
				}
				
				free(e);
			}
		}
		
		m->rows = realloc(m->rows, sizeof(void*)*n_rows);
		
		m->n_rows = n_rows;
	}
	
	if (n_cols > m->n_cols)
	{
		/* All that we need to do is expand the column array */
		m->cols = realloc(m->cols, sizeof(void*)*n_cols);
		bzero(m->cols+m->n_cols, sizeof(void*)*(n_cols - m->n_cols));
		
		m->n_cols = n_cols;
	}
	else if (n_cols < m->n_cols)
	{
		/* Cut off the rows at the new boundary */
		int row;
		for (row=0; row<m->n_rows; row++)
		{
			spm_matrix_element *e;
			for (e=m->rows[row]; e; e=e->next_in_row)
			{
				if (e->next_in_row && e->next_in_row->col >= n_cols)
				{
					e->next_in_row = NULL;
					break;
				}
			}
		}
		
		/* Free the columns which no longer exist */
		int col;
		for (col=m->n_cols; col<n_cols; col++)
		{
			spm_matrix_element *e, *e_next;
			for (e=m->cols[col]; e; e=e_next)
			{
				e_next = e->next_in_col;
				
				if (m->vinfo->free_func)
				{
					m->vinfo->free_func(m->vinfo->udata, SPM_M_VALUE(e));
				}
				
				free(e);
			}
		}
		
		m->cols = realloc(m->cols, sizeof(void*)*n_cols);
		
		m->n_cols = n_cols;
	}
}

/* THIS FUNCTION IS CURRENTLY BROKEN */
void
spm_matrix_transfer_cols(spm_matrix *src, int start, int count, spm_matrix *dest, int dstart)
{
	int i, n_rows = src->n_rows;
	
	assert(src!=dest);
	assert(src->n_rows == dest->n_rows);
	assert(src->vinfo == dest->vinfo);
	assert(start+count <= src->n_cols);
	assert(dstart+count <= dest->n_cols);
	
	/* Transfer the row pointers while simultaneously "sewing up" the gap left by the missing columns */
	for (i=0; i<n_rows; i++)
	{
		spm_matrix_element **left_end = &src->rows[i];
		while (*left_end && (*left_end)->col<start)
		{
			left_end = &(*left_end)->next_in_row;
		}
		
		spm_matrix_element **right_end = left_end;
		while (*right_end && (*right_end)->col<start+count)
		{
			right_end = &(*right_end)->next_in_row;
		}
		
		dest->rows[i] = *left_end;
		*left_end = *right_end;
		*right_end = NULL;
	}
	
	/* Transfer the pointers to the beginnings of the columns */
	memcpy(dest->cols+dstart, src->cols+start, sizeof(void*)*count);
	bzero(src->cols+start, sizeof(void*)*count);
	
	/* Renumber the elements transferred if necessary */
	if (start!=dstart)
	{
		for (i=0; i<n_rows; i++)
		{
			spm_matrix_element *e;
			for (e=dest->rows[i]; e; e=e->next_in_row)
			{
				e->col -= dstart-start;
			}
		}
	}
}

void
spm_matrix_del_row(spm_matrix *m, int row)
{
	spm_matrix_element **slot, *e;
	int col;
	for (col=0; col<m->n_cols; col++)
	{
		/* Find the place where the row to be removed is, or where it would be */
		for (slot=&m->cols[col]; *slot && (*slot)->row<row; slot=&(*slot)->next_in_col)
		{
			continue;
		}
		
		/* If the row to be removed has an element in this column, free it */
		if ((*slot)->row==row)
		{
			spm_matrix_element *to_free = *slot;
			if (m->vinfo->free_func)
			{
				m->vinfo->free_func(m->vinfo->udata, SPM_V_VALUE(to_free));
			}
			*slot = to_free->next_in_col;
		}
		
		/* Update the row numbers of the rows below the one to be removed */
		for (e=*slot; e; e=e->next_in_col)
		{
			e->row--;
		}
	}
	
	/* Update the row pointers */
	memmove(m->rows+row, m->rows+row+1, sizeof(void*)*(m->n_rows-(row+1)));
	
	m->n_rows--;
}

void
spm_matrix_del_col(spm_matrix *m, int col)
{
	spm_matrix_element **slot, *e;
	int row;
	for (row=0; row<m->n_rows; row++)
	{
		/* Find the place where the col to be removed is, or where it would be */
		for (slot=&m->rows[row]; *slot && (*slot)->col<col; slot=&(*slot)->next_in_row)
		{
			continue;
		}
		
		/* If the column to be removed has an element in this row, free it */
		if ((*slot)->col==col)
		{
			spm_matrix_element *to_free = *slot;
			if (m->vinfo->free_func)
			{
				m->vinfo->free_func(m->vinfo->udata, SPM_V_VALUE(to_free));
			}
			*slot = to_free->next_in_row;
		}
		
		/* Update the col numbers of the columns to the right of the one to be removed */
		for (e=*slot; e; e=e->next_in_row)
		{
			e->col--;
		}
	}
	
	/* Update the col pointers */
	memmove(m->cols+col, m->cols+col+1, sizeof(void*)*(m->n_cols-(col+1)));
	
	m->n_cols--;
}

void
spm_matrix_print_pointers(spm_matrix *m)
{	
	printf("Matrix %p - %d rows, %d columns:\n", m, m->n_rows, m->n_cols);
	printf("Printing element pointers.\n");
	int row;
	for (row=0; row<m->n_rows; row++)
	{
		printf("Row %d:", row);
		
		spm_matrix_element *e;
		for (e=m->rows[row]; e; e=e->next_in_row)
		{
			printf("\t[%d]%p", e->col, e); 
		}
		
		printf("\n");
	}
	printf("End matrix.\n");
}