#include "spm.h"
#include "spm_multiply.h"
#include "spm_gaussian.h"
#include "spm_value_double.h"
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <math.h>

static void
double_set_zero_func(void *udata, void *buffer)
{
	*(double*)buffer = 0;
}

static void
double_set_one_func(void *udata, void *buffer)
{
	*(double*)buffer = 1;
}

static int
double_is_zero_func(void *udata, const void *buffer)
{
	return *(double*)buffer == 0;
}

static void
double_add_func(void *udata, void *buf_in1, void *buf_in2, void *buf_out)
{
	double a = *(double*)buf_in1;
	double b = *(double*)buf_in2;
	*(double*)buf_out = a+b;
}

static void
double_subtract_func(void *udata, void *buf_in1, void *buf_in2, void *buf_out)
{
	double a = *(double*)buf_in1;
	double b = *(double*)buf_in2;
	*(double*)buf_out = a-b;
}

static void
double_multiply_func(void *udata, void *buf_in1, void *buf_in2, void *buf_out)
{
	double a = *(double*)buf_in1;
	double b = *(double*)buf_in2;
	*(double*)buf_out = a*b;
}

static void
double_divide_func(void *udata, void *buf_in1, void *buf_in2, void *buf_out)
{
	double a = *(double*)buf_in1;
	double b = *(double*)buf_in2;
	*(double*)buf_out = a/b;
}

/* The smallest permitted pivot is 10.0^(-10). PIVOT_TOLERANCE is the log (base FLT_RADIX) of the
smallest permitted pivot. */
#if FLT_RADIX==2
#define PIVOT_TOLERANCE -33
#else
#define PIVOT_TOLERANCE (-10*logb(10))
#endif

static int
double_pivot_check_func(void *udata, const void *buf)
{
	double pivot = *(double*)buf;
	int exponent = ilogb(pivot);
	if (exponent <= PIVOT_TOLERANCE)
	{
		return -1;
	}
	else
	{
		return exponent - PIVOT_TOLERANCE;
	}
}

const spm_value_info 
spm_double_value_info =
{
	sizeof(double),
	double_set_zero_func,
	double_set_one_func,
	double_is_zero_func,
	NULL,
	NULL
};

const spm_multiply_info
spm_double_multiply_info =
{
	&spm_double_value_info,
	&spm_double_value_info,
	&spm_double_value_info,
	double_multiply_func,
	double_add_func,
	NULL
};

const spm_gaussian_info
spm_double_gaussian_info =
{
	&spm_double_value_info,
	double_divide_func,
	double_multiply_func,
	double_subtract_func,
	double_pivot_check_func,
	NULL
};

const spm_gaussianaugmentation_info
spm_double_gaussianaugmentation_info =
{
	&spm_double_value_info,
	&spm_double_value_info,
	double_divide_func,
	double_multiply_func,
	double_subtract_func,
	NULL
};

spm_matrix *
spm_matrix_double_alloc(int n_rows, int n_cols)
{
	return spm_matrix_alloc(n_rows, n_cols, &spm_double_value_info);
}

void
spm_matrix_double_set(spm_matrix *m, int row, int col, double value)
{
	assert(m->vinfo == &spm_double_value_info);
	spm_matrix_set(m, row, col, &value);
}

double
spm_matrix_double_get(spm_matrix *m, int row, int col)
{
	assert(m->vinfo == &spm_double_value_info);
	double value;
	spm_matrix_get(m, row, col, &value);
	return value;
}

void
spm_matrix_double_print(spm_matrix *m)
{
	spm_matrix_double_write(m, stdout);
}

void
spm_matrix_double_write(spm_matrix *m, FILE *f)
{
	assert(m);
	assert(m->vinfo == &spm_double_value_info);
	assert(f);
	
	fprintf(f, "Matrix %p: %d x %d\n", m, m->n_rows, m->n_cols);
	int row;
	for (row=0; row<m->n_rows; row++)
	{
		fprintf(f, "Row %d:", row);
		
		spm_matrix_element *e;
		for (e=m->rows[row]; e; e=e->next_in_row)
		{
			fprintf(f, "\t%d=%lf", e->col, *(double*)SPM_M_VALUE(e)); 
		}
		
		fprintf(f, "\n");
	}
	fprintf(f,  "End matrix.\n");
}